# -*- coding: utf-8 -*-
"""provides sequencing fetching from NCBI and Ensembl

"""

import logging
import os

import bioutils.seqfetcher

from ..exceptions import HGVSDataNotAvailableError

_logger = logging.getLogger(__name__)


class SeqFetcher:
    """This class is intended primarily as a mixin for HGVS data providers
    that doen't otherwise have access to sequence data.  It uses the
    fetch_seq() function in this module to fetch sequences from
    several sources; see that function for details.

    >> sf = SeqFetcher()

    >> sf.fetch_seq('NP_056374.2',0,10)
    'MESRETLSSS'

    """

    def __init__(self):
        # If HGVS_SEQREPO_DIR is defined, we use seqrepo for *all* sequences.
        # If HGVS_SEQREPO_URL is defined, use instance of seqrepo-rest-service for *all* sequences.
        #   (see https://github.com/biocommons/seqrepo-rest-service for more info)
        # Otherwise, we fall back to remote sequence fetching
        seqrepo_dir = os.environ.get("HGVS_SEQREPO_DIR")
        seqrepo_url = os.environ.get("HGVS_SEQREPO_URL")
        if seqrepo_dir:
            from biocommons.seqrepo import SeqRepo

            self.sr = SeqRepo(seqrepo_dir)

            def _fetch_seq_seqrepo(ac, start_i=None, end_i=None):
                return self.sr.fetch(ac, start_i, end_i)

            self.fetcher = _fetch_seq_seqrepo
            self.source = "SeqRepo ({})".format(seqrepo_dir)
        elif seqrepo_url:
            from biocommons.seqrepo.dataproxy import SeqRepoRESTDataProxy

            self.sr = SeqRepoRESTDataProxy(seqrepo_url)
            self.fetcher = lambda ac, start_i=None, end_i=None: self.sr.get_sequence(
                ac, start_i, end_i
            )
            self.source = f"SeqRepo REST ({seqrepo_url})"
        else:
            self.sr = None
            self.fetcher = bioutils.seqfetcher.fetch_seq
            self.source = "bioutils.seqfetcher (network fetching)"
        _logger.info("Fetching sequences with " + self.source)

    def fetch_seq(self, ac, start_i=None, end_i=None):
        try:
            return self.fetcher(ac, start_i, end_i)
        except Exception as ex:
            raise HGVSDataNotAvailableError(
                "Failed to fetch {ac} from {self.source} ({ex})".format(ac=ac, ex=ex, self=self)
            )


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



"""provides CIGAR-based mapping for hgvs

Heads up: UTA CIGAR strings are relative to the transcript; that is,
the transcript is the reference and the genome is the target sequence.
hgvs, and this module specifically, swaps this nomenclature so that
the genome is ref and the transcript is the target.

UTA CIGAR strings are relative to the transcript.  Specifically:

  * D is a deletion in the genome relative to the transcript
  * I is an insertion in the genome relative to the transcript
  * N is an intronic insertion relative to the transcript

The conceptual swap occurs in the _map function.

"""


import re

from hgvs.exceptions import HGVSInvalidIntervalError


class CIGARMapper:
    """provides coordinate mapping between two sequences whose alignment
    is given by a CIGAR string

    CIGAR is about alignments between positions in two sequences.  It
    is base-centric.

    Unfortunately, base-centric coordinate systems require additional
    complexity to refer to zero-width positions.

    This code uses interbase intervals.  Interbase positions are
    zero-width boundaries between bases.  They often look similar to
    zero-based, right open coordinates. (But don't call them that.  It
    upsets me deeply.)  The most important difference is that zero
    width intervals neatly represent insertions between bases (or
    before or after the sequence).

    """

    def __init__(self, cigar):
        self.cigar = cigar
        self.ref_pos, self.tgt_pos, self.cigar_op = _parse_cigar(self.cigar)

    def __repr__(self):
        return f"CIGARMapper(cigar='{self.cigar}')"

    @property
    def ref_len(self):
        return self.ref_pos[-1]

    @property
    def tgt_len(self):
        return self.tgt_pos[-1]

    def map_ref_to_tgt(self, pos, end, strict_bounds=True):
        return self._map(
            from_pos=self.ref_pos,
            to_pos=self.tgt_pos,
            pos=pos,
            end=end,
            strict_bounds=strict_bounds,
        )

    def map_tgt_to_ref(self, pos, end, strict_bounds=True):
        return self._map(
            from_pos=self.tgt_pos,
            to_pos=self.ref_pos,
            pos=pos,
            end=end,
            strict_bounds=strict_bounds,
        )

    def _map(self, from_pos, to_pos, pos, end, strict_bounds):
        """Map position between aligned segments

        Positions in this function are 0-based, base-counting.
        """

        if strict_bounds and (pos < 0 or pos > from_pos[-1]):
            raise HGVSInvalidIntervalError("Position is beyond the bounds of transcript record")

        # find aligned segment to use as basis for mapping
        # okay for pos to be before first element or after last
        for pos_i in range(len(self.cigar_op)):
            if pos < from_pos[pos_i + 1]:
                break

        if self.cigar_op[pos_i] in "=MX":
            mapped_pos = to_pos[pos_i] + (pos - from_pos[pos_i])
            mapped_pos_offset = 0

        elif self.cigar_op[pos_i] in "DI":
            mapped_pos = to_pos[pos_i]
            if end == "start":
                mapped_pos -= 1
            mapped_pos_offset = 0

        elif self.cigar_op[pos_i] == "N":
            if pos - from_pos[pos_i] + 1 <= from_pos[pos_i + 1] - pos:
                mapped_pos = to_pos[pos_i] - 1
                mapped_pos_offset = pos - from_pos[pos_i] + 1
            else:
                mapped_pos = to_pos[pos_i]
                mapped_pos_offset = -(from_pos[pos_i + 1] - pos)

        return mapped_pos, mapped_pos_offset, self.cigar_op[pos_i]


def _parse_cigar(cigar):
    """For a given CIGAR string, return the start positions of each
    aligned segment in ref and tgt, and a list of CIGAR operators.

    """

    cigar_re = re.compile(r"(?P<len>\d+)?(?P<op>[=DIMNX])")
    advance_ref_ops = "=MXIN"
    advance_tgt_ops = "=MXD"

    ces = [m.groupdict() for m in cigar_re.finditer(cigar)]
    cigar_len = len(ces)

    ref_pos = [None] * cigar_len
    tgt_pos = [None] * cigar_len
    cigar_op = [None] * cigar_len
    ref_cur = tgt_cur = 0
    for i, ce in enumerate(ces):
        ref_pos[i] = ref_cur
        tgt_pos[i] = tgt_cur
        cigar_op[i] = ce["op"]
        step = int(ce["len"] or 1)
        # Note: these cases are
        if ce["op"] in advance_ref_ops:
            ref_cur += step
        if ce["op"] in advance_tgt_ops:
            tgt_cur += step
    ref_pos.append(ref_cur)
    tgt_pos.append(tgt_cur)
    return ref_pos, tgt_pos, cigar_op



# -*- coding: utf-8 -*-
"""Internal utility to display textual representation of variant context

This code requires uta-align and pysam packages, which are NOT part of
a typical hgvs installation because they change periodically in ways
that break dependencies.


Something like this:

(default-2.7)snafu$ ./sbin/hgvs-shell
In [1]: import hgvs.utils.context as huc
In [2]: h = 'NM_000399.3:n.2689_2690insA'
In [3]: v = hp.parse_hgvs_variant(h)
In [4]: vn = hn.normalize(v)
In [5]: print(huc.variant_context_w_alignment(am, vn))
                                              v                               NC_000010.10:g.64572045dupT
NC_000010.10 g 64572025 > ACTCAGGGAGTGATTTTTTTTCTCCATAATAAGGCAACCCA          > 64572065 NC_000010.10:g.64572045dupT
NC_000010.10 g 64572025 < TGAGTCCCTCACTAAAAAAAAGAGGTATTATTCCGTTGGGT          < 64572065 NC_000010.10:g.64572045dupT
                          |||||||||||||-|||||||||||||||||||||||||||          13=1D27=
NM_000399.3  n     2670 < TGAGTCCCTCACT-AAAAAAAGAGGTATTATTCCGTTGGGT          <     2709 NM_000399.3:n.2696dupA
NM_000399.3  c      902 <                                                    <      941 NM_000399.3:c.*928dupA

"""

import re

from bioutils.sequences import complement

from ..location import Interval, SimplePosition


def full_house(am, var, tx_ac=None):
    if var.type == "g":
        var_g = var
        if tx_ac is None:
            rtx = am.relevant_transcripts(var)
            if len(rtx) == 0:
                raise RuntimeError("no relevant transcripts for {var.ac}".format(var=var))
            if len(rtx) > 1:
                raise RuntimeError(
                    "{n} relevant transcripts for {var.ac}; you need to pick one".format(
                        n=len(rtx), var=var
                    )
                )
            tx_ac = rtx[0]
        var_n = am.g_to_n(var_g, tx_ac)
        var_c = am.n_to_c(var_n)
    elif var.type == "n":
        var_n = var
        var_g = am.n_to_g(var_n)
        var_c = am.n_to_c(var_n)
    elif var.type == "c":
        var_c = var
        var_g = am.c_to_g(var_c)
        var_n = am.c_to_n(var_c)
    var_p = am.c_to_p(var_c)
    return {"g": var_g, "c": var_c, "n": var_n, "p": var_p}


# def variant_context(am, var, margin=20):
#     span = _span(var, margin)
#     span_g = _ival_to_span(fh['g'])
#     span_g[0] -= margin
#     span_g[1] += margin
#     return '\n'.join([
#         seq_line_fmt(var=var,
#                      span=span,
#                      content=am.hdp.get_seq(var.ac, *span),
#                      post=''), pointer_line(var, span)
#     ])


def variant_context_w_alignment(am, var, margin=20, tx_ac=None):
    """This module is experimental. It requires the uta_align package from pypi."""
    from uta_align.align.algorithms import align, cigar_alignment

    fh = full_house(am, var, tx_ac=tx_ac)
    tm = am._fetch_AlignmentMapper(fh["n"].ac, fh["g"].ac, am.alt_aln_method)
    strand = tm.strand
    span_g = _ival_to_span(fh["g"].posedit.pos)
    span_g = (span_g[0] - margin, span_g[1] + margin)
    ival_g = Interval(SimplePosition(span_g[0]), SimplePosition(span_g[1]))
    ival_n = tm.g_to_n(ival_g)
    assert ival_n.start.offset == 0 and ival_n.end.offset == 0, "limited to coding variants"
    span_n = _ival_to_span(ival_n)
    ival_c = tm.g_to_c(ival_g)
    span_c = _ival_to_span(ival_c)
    seq_gt = am.hdp.get_seq(fh["g"].ac, span_g[0] - 1, span_g[1])
    seq_gb = complement(seq_gt)
    seq_n = am.hdp.get_seq(fh["n"].ac, span_n[0] - 1, span_n[1])
    if strand == 1:
        a = align(bytes(seq_gt), bytes(seq_n), b"global", extended_cigar=True)
    else:
        seq_n = "".join(reversed(seq_n))
        a = align(bytes(seq_gb), bytes(seq_n), b"global", extended_cigar=True)

    aseq_gt, _ = cigar_alignment(seq_gt, a.query, a.cigar, hide_match=False)
    aseq_gb, aseq_n = cigar_alignment(seq_gb, a.query, a.cigar, hide_match=False)
    aln_str = _reformat_aln_str(cigar_alignment(a.ref, a.query, a.cigar, hide_match=True)[1])
    s_dir = ">" if strand == 1 else "<"

    lines = [
        [
            1,
            0,
            seq_line_fmt(
                var=fh["c"],
                span=span_c if strand == 1 else list(reversed(span_c)),
                content="",
                dir=s_dir,
            ),
        ],
        [
            2,
            0,
            seq_line_fmt(
                var=fh["n"],
                span=span_n if strand == 1 else list(reversed(span_n)),
                content=aseq_n,
                dir=s_dir,
            ),
        ],
        [
            3,
            0,
            _line_fmt.format(pre="", content=aln_str, post=a.cigar.to_string(), comment=""),
        ],
        [
            4,
            1,
            seq_line_fmt(var=fh["g"], span=span_g, content=aseq_gt, dir=">"),
        ],
        [
            4,
            2,
            seq_line_fmt(var=fh["g"], span=span_g, content=aseq_gb, dir="<"),
        ],
        [
            5,
            0,
            pointer_line(var=fh["g"], span=span_g),
        ],
    ]
    if strand == -1:
        lines.sort(key=lambda e: (-e[0], e[1]))
    return "\n".join(r[2] for r in lines)


def _ival_to_span(ival):
    return (ival.start.base, ival.end.base)


def _reformat_aln_str(aln_str):
    return re.sub(r"[ACGT]", " ", aln_str.replace(".", "|"))


# pre=[ac c s] d content d post=[end] comment
_line_fmt = "{pre:>30s} {content:45s} {post} {comment}"
_pre_fmt = "{ac:12s} {type:1s} {s:10d} {dir:1s}"
_post_fmt = "{dir:1s} {e:8d}"


def seq_line_fmt(var, span, content, dir=""):
    return _line_fmt.format(
        pre=_pre_fmt.format(ac=var.ac, type=var.type, s=span[0], dir=dir),
        content=content,
        post=_post_fmt.format(dir=dir, e=span[1]),
        comment=str(var),
    )


def pointer_line(var, span):
    s0 = span[0]
    o = var.posedit.pos.start.base - s0
    l = var.posedit.pos.end.base - var.posedit.pos.start.base + 1  # noqa: E741
    if var.posedit.edit.type == "ins":
        p = " " * o + "><"
    else:
        p = " " * o + "*" * l
    return _line_fmt.format(pre="", content=p, post="", comment=str(var))


def format_sequence(seq, start=None, end=None, group_size=3):
    """print seq from [start, end) in groups of size

           3   6   9  12  15
           |   |   |   |   |
    2001 AAA BBB CCC DDD EEE

    """

    width = 100
    loc_width = 9
    sep = " "
    body_sep = " : "

    start = start or 0
    end = end or len(seq)

    bw = width - loc_width - len(body_sep)
    assert group_size <= bw, "group size must be less than available line width"
    gpl = int((bw + len(sep)) / (group_size + len(sep)))  # groups per line
    gpl = int(gpl / 5) * 5 if gpl > 20 else gpl
    rpl = group_size * gpl
    line_fmt = "{{l:>{lw}s}}{body_sep}{{body}}".format(lw=loc_width, body_sep=body_sep)
    ge_fmt = "{{ge:>{gs}}}".format(gs=group_size)

    blocks = []
    for ls in range(start, end, rpl):
        le = ls + rpl

        groups = [
            ge_fmt.format(ge=str(gs + group_size)[-group_size + 1 :])
            for gs in range(ls, le, group_size)
        ]
        blocks += [line_fmt.format(l="", body=sep.join(groups)) + "\n"]

        groups = [seq[gs : min(gs + group_size, end)] for gs in range(ls, le, group_size)]
        blocks += [line_fmt.format(l=str(ls + 1), body=sep.join(groups)) + "\n"]

        blocks += ["\n"]

    return blocks



# -*- coding: utf-8 -*-
import csv
import os
import re
import unittest

import hgvs.dataproviders.uta
import hgvs.parser
import hgvs.variantmapper
from support import CACHE


# Real data - cp tests


def gcp_file_reader(fn):
    rdr = csv.DictReader(open(fn, "r"), delimiter=str("\t"))
    for rec in rdr:
        if rec["id"].startswith("#"):
            continue
        yield rec


class TestHgvsCToPReal(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hdp = hgvs.dataproviders.uta.connect(
            mode=os.environ.get("HGVS_CACHE_MODE", "learn"), cache=CACHE
        )
        cls._hm = hgvs.variantmapper.VariantMapper(cls.hdp)
        cls._hp = hgvs.parser.Parser()
        cls._failed = []

    def test_c_to_p_ext(self):
        infilename = "ext.tsv"
        outfilename = "ext.out"
        infile = os.path.join(os.path.dirname(__file__), "data", infilename)
        outfile = os.path.join(os.path.dirname(__file__), "data", outfilename)
        self._run_cp_test(infile, outfile)

    #
    # internal methods
    #

    def _run_cp_test(self, infile, outfile):
        with open(outfile, "w") as out:
            out.write("id\tHGVSg\tHGVSc\tHGVSp\tConverterResult\tError\n")
            self._dup_regex = re.compile(r"dup[0-9]+$")
            for rec in gcp_file_reader(infile):
                self._run_comparison(rec, out)
        msg = "# failed: {}".format(len(self._failed))
        self.assertTrue(len(self._failed) == 0, msg)

    def _run_comparison(self, rec, out):
        (row_id, hgvsg, hgvsc, hgvsp_expected) = (
            rec["id"],
            rec["HGVSg"],
            rec["HGVSc"],
            rec["HGVSp"],
        )

        hgvsc = self._dup_regex.sub("dup", hgvsc)  # cleanup dupN

        if not row_id.startswith("#") and hgvsc and hgvsp_expected:
            try:
                var_c = self._hp.parse_hgvs_variant(hgvsc)
                var_p = self._hm.c_to_p(
                    var_c, hgvsp_expected.split(":")[0]
                )  # hack until p.?, p.= etc parse
                hgvsp_actual = str(var_p)

                if hgvsp_expected != hgvsp_actual:
                    if hgvsp_expected.replace("*", "Ter") == hgvsp_actual:
                        pass
                    else:
                        self._append_fail(
                            out, row_id, hgvsg, hgvsc, hgvsp_expected, hgvsp_actual, "MISMATCH"
                        )
            except Exception as e:
                self._append_fail(
                    out, row_id, hgvsg, hgvsc, hgvsp_expected, "NO_OUTPUT_EXCEPTION", e.message
                )

    def _append_fail(self, out, row_id, hgvsg, hgvsc, hgvsp_expected, hgvsp_actual, msg):
        self._failed.append((row_id, hgvsg, hgvsc, hgvsp_expected, hgvsp_actual, msg))
        out.write(
            "{}\t{}\t{}\t{}\t{}\t{}\n".format(
                row_id, hgvsg, hgvsc, hgvsp_expected, hgvsp_actual, msg
            )
        )

    def test_c_to_p_format(self):
        hgvsc = "NM_022464.4:c.3G>A"
        hgvsp_expected = "NP_071909.1:p.Met1?"
        hgvsp_expected_alternative = "NP_071909.1:p.?"

        var_c = self._hp.parse_hgvs_variant(hgvsc)
        var_p = self._hm.c_to_p(var_c, hgvsp_expected.split(":")[0])

        default_format_p = var_p.format()
        self.assertEqual(hgvsp_expected, default_format_p)

        alt_format_p = var_p.format(conf={"p_init_met": False})
        self.assertEqual(hgvsp_expected_alternative, alt_format_p)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
"""defined exceptions used by the hgvs package

"""


class HGVSError(Exception):
    pass


class HGVSDataNotAvailableError(HGVSError):
    pass


class HGVSInternalError(HGVSError):
    pass


class HGVSInvalidIntervalError(HGVSError):
    pass


class HGVSInvalidVariantError(HGVSError):
    """Exception raised when variant is inconsistent or invalid"""

    pass


class HGVSNormalizationError(HGVSError):
    pass


class HGVSParseError(HGVSError):
    pass


class HGVSUnsupportedOperationError(HGVSError):
    pass


class HGVSUsageError(HGVSError):
    """Exception raised when client/caller has made an invalid request"""

    pass


class HGVSVerifyFailedError(HGVSError):
    """Exception raised when the cached hdp result is not consistent with latest remote result"""

    pass


# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



# -*- coding: utf-8 -*-
import unittest

import pytest

import hgvs.location
import hgvs.parser
from hgvs.enums import Datum
from hgvs.exceptions import HGVSError, HGVSUnsupportedOperationError


@pytest.mark.quick
@pytest.mark.models
class Test_SimplePosition(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()

    def test_success(self):
        self.assertEqual(str(hgvs.location.SimplePosition(5)), "5")
        self.assertEqual(str(hgvs.location.SimplePosition(5, uncertain=True)), "(5)")
        self.assertEqual(str(hgvs.location.SimplePosition(None)), "?")

    def test_failure(self):
        with self.assertRaises(AssertionError):
            self.assertEqual(hgvs.location.SimplePosition(-1), "SHOULD FAIL")

    def test_simple_subtraction(self):
        self.assertEqual(hgvs.location.SimplePosition(5) - hgvs.location.SimplePosition(3), 2)

    def test_simple_comparision(self):
        var = self.hp.parse_hgvs_variant("NC_000007.13:g.36561662_36561683del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NC_000007.13:g.36561662C>T")
        self.assertTrue(var.posedit.pos.start == var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start >= var.posedit.pos.end)


@pytest.mark.quick
class Test_BaseOffsetPosition(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()

    def test_success(self):
        # r.5
        cdsp = hgvs.location.BaseOffsetPosition(5)
        self.assertEqual(cdsp.datum, Datum.SEQ_START)
        self.assertEqual(cdsp.base, 5)
        self.assertEqual(cdsp.offset, 0)
        self.assertEqual(str(cdsp), "5")
        self.assertFalse(cdsp.is_intronic)

        # r.5+6
        cdsp.offset = 6
        self.assertEqual(str(cdsp), "5+6")
        self.assertTrue(cdsp.is_intronic)

        # r.5+?
        cdsp.offset = None
        self.assertEqual(str(cdsp), "5+?")
        self.assertTrue(cdsp.is_intronic)

        # r.(5+?)
        cdsp.uncertain = True
        self.assertEqual(str(cdsp), "(5+?)")

        # c.*5
        cdsp = hgvs.location.BaseOffsetPosition(5, datum=Datum.CDS_END)
        self.assertEqual(cdsp.datum, Datum.CDS_END)
        self.assertEqual(cdsp.base, 5)
        self.assertEqual(cdsp.offset, 0)
        self.assertEqual(str(cdsp), "*5")

        cdsp.uncertain = True
        self.assertEqual(str(cdsp), "(*5)")

        cdsp.offset = 7
        self.assertEqual(str(cdsp), "(*5+7)")

    def test_baseoffset_subtraction(self):
        v30 = hgvs.location.BaseOffsetPosition(3, 0)
        v50 = hgvs.location.BaseOffsetPosition(5, 0)
        v54 = hgvs.location.BaseOffsetPosition(5, 4)

        self.assertEqual(v50 - v30, 2)

        with self.assertRaises(HGVSError):
            _ = v54 - v30

    def test_baseoffset_comparision(self):
        var = self.hp.parse_hgvs_variant("NM_000030.2:c.669_680del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.679_680+2del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.-6_680+2del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.680+2_680+10del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.680+2_*82del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.-12_*82del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.680+2_681del")
        self.assertFalse(var.posedit.pos.start == var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertTrue(var.posedit.pos.start <= var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start >= var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NM_000030.2:c.680+2_681-32del")
        with self.assertRaises(HGVSUnsupportedOperationError):
            var.posedit.pos.start < var.posedit.pos.end


@pytest.mark.quick
class Test_AAPosition(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.hp = hgvs.parser.Parser()

    def test_AAPosition(self):
        ap = hgvs.location.AAPosition(15, "S")
        self.assertEqual(ap.pos, 15)
        self.assertEqual(str(ap), "Ser15")

    def test_aaposition_subtraction(self):
        l1 = hgvs.location.AAPosition(15, "S")
        l2 = hgvs.location.AAPosition(20, "S")
        self.assertEqual(l2 - l1, 5)

    def test_aaposition_comparision(self):
        var = self.hp.parse_hgvs_variant("NP_000042.3:p.His1082_Val1085delinsLeuHisGlnAla")
        self.assertTrue(var.posedit.pos.start < var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)

        var = self.hp.parse_hgvs_variant("NP_000042.3:p.His1082ArgfsTer2")
        self.assertFalse(var.posedit.pos.start < var.posedit.pos.end)
        self.assertFalse(var.posedit.pos.start > var.posedit.pos.end)


@pytest.mark.quick
class Test_Interval(unittest.TestCase):
    def test_Interval(self):
        ival = hgvs.location.Interval(
            hgvs.location.BaseOffsetPosition(base=12, offset=+34),
            hgvs.location.BaseOffsetPosition(base=56, offset=-78),
        )
        self.assertEqual(ival.start.base, 12)
        self.assertEqual(ival.start.offset, 34)
        self.assertEqual(ival.end.base, 56)
        self.assertEqual(ival.end.offset, -78)
        self.assertEqual(str(ival), "12+34_56-78")

    def test_length(self):
        ival = hgvs.location.Interval(
            hgvs.location.BaseOffsetPosition(base=12, offset=0),
            hgvs.location.BaseOffsetPosition(base=50, offset=0),
        )
        self.assertEqual(ival._length(), 39)


if __name__ == "__main__":
    unittest.main()

# <LICENSE>
# Copyright 2018 HGVS Contributors (https://github.com/biocommons/hgvs)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# </LICENSE>



import math

from hgvs.pretty.models import VariantData
from hgvs.pretty.console.renderer import BasicRenderer


class TxAligRenderer(BasicRenderer):
    """
    TxAligRenderer is a class that extends BasicRenderer to provide functionality for rendering
    the transcript sequence, optional with alternating colors per codon.
    It includes methods to generate a legend and display data with the transcript sequence.

    Methods
    -------
    legend() -> str:
        Generates a legend string indicating the orientation of the transcript sequence.
    display(data: VariantData) -> str:
        Generates a string representation of the variant data, showing details of the transcript
        for the specified region. It uses color coding if enabled in the configuration.
    """

    def legend(self) -> str:
        if self.orientation > 0:
            orientation = "->"
        elif self.orientation < 0 and self.config.reverse_display:
            orientation = "->"
        else:
            orientation = "<-"
        return f"tx seq {orientation} : "

    def display(self, data: VariantData) -> str:
        """If transcript info is available show the details of the tx for the region."""
        if not data.var_c_or_n:
            return ""

        from hgvs.pretty.console.constants import ENDC, COLOR_MAP

        var_str = ""

        first_c = None
        last_c = None

        coding = False
        c_pos = None

        counter = -1
        for pdata in data.position_details:
            counter += 1

            if pdata.mapped_pos is None:
                var_str += " "
                continue

            n_pos = pdata.n_pos
            c_pos = pdata.c_pos
            cig = pdata.cigar_ref
            c_offset = pdata.c_offset

            if cig == "N" or (c_offset is not None and c_offset != 0):
                coding = False
                var_str += " "
                continue

            if c_pos:
                # if we get here we are coding...
                last_c = f"c.{c_pos}"
            else:
                last_c = f"n.{n_pos}"
                c_pos = n_pos
            if not coding:
                if not first_c:
                    first_c = last_c

                coding = True

            if cig == "=":
                if c_pos:
                    bg_col = math.ceil((c_pos) / 3) % 2
                elif n_pos:
                    bg_col = math.ceil((n_pos) / 3) % 2
                else:
                    var_str += " "
                    continue

                # print(p, c_pos, c3, bg_col )
                if n_pos >= 0:
                    base = pdata.tx
                    if self.config.use_color and c_pos > 0:
                        if bg_col:
                            var_str += COLOR_MAP["codon1"] + base + ENDC
                        else:
                            var_str += COLOR_MAP["codon2"] + base + ENDC
                    else:
                        if c_pos is None or c_pos < 0:
                            var_str += base.lower()
                        else:
                            var_str += base
                    continue

            elif cig == "X" or cig == "D":
                # for mismatches and tx-insertions show sequence
                var_str += pdata.tx
                continue

            else:
                var_str += "-"

        return var_str



# -*- coding: utf-8 -*-
""" A class to manage conversion of SequenceVariants to repeat representation"""

from dataclasses import dataclass
import re
import hgvs
from hgvs.pretty.models import VariantCoords


@dataclass(eq=True, repr=True, frozen=True, order=True)
class RepeatUnit:
    repeat_count: int
    repeat_unit:str
    block_size:int
    block: str


class RepeatAnalyser:

    def __init__(self, fs: VariantCoords, reverse:bool=False) -> None:

        self.is_repeat = False
        self.repeat_units_ref = None
        self.repeat_units_alt = None
        self.ref_str = fs.ref
        self.alt_str = fs.alt
        self.reverse = reverse

        # protect from too large sequences getting analyzed.
        if len(self.ref_str) > hgvs.global_config.repeats.max_repeat_length:
            self.repeat_units_ref = []
            self.repeat_units_alt = []
            return

        self.repeat_units_ref = detect_repetitive_block_lengths(self.ref_str, reverse=self.reverse)
        self.repeat_units_alt = detect_repetitive_block_lengths(self.alt_str, reverse = self.reverse)

        if len(self.repeat_units_ref) == 0 and len(self.repeat_units_alt) ==0 :
            return
                
        # check longest repeat blocks:
        # we only look at ref to determine if there are repeats 
        # If ref has no repeat, we don't call this a repeat variant, even if alt would have a repetitive unit
        longest_r_unit = self._get_longest_repeat_unit(self.repeat_units_ref)
        if longest_r_unit is None:
            return
            
        # filter our too fragmented results
        expected_size = len(self.ref_str) / 3
        if longest_r_unit.block_size < expected_size:            
            return
            
        if longest_r_unit.repeat_unit not in self.alt_str:
            return

        self.repeat_units_alt = detect_repetitive_block_lengths(self.alt_str, longest_ref_unit = longest_r_unit, reverse = self.reverse)

        self.is_repeat = True

        ref_repeat = get_repeat_str(self.ref_str, self.alt_str, self.repeat_units_ref, self.repeat_units_alt, self.reverse)
        alt_repeat = get_repeat_str(self.alt_str, self.ref_str, self.repeat_units_alt, self.repeat_units_ref, self.reverse)
        self.ref_str = ref_repeat
        self.alt_str = alt_repeat

    def __repr__(self):
        return f"{self.ref_str}>{self.alt_str}"

    def _get_longest_repeat_unit(self, repeat_units:list[RepeatUnit])->RepeatUnit:
        lru = None
        for ru in repeat_units:
            if not lru:
                lru = ru
                continue

            if ru.block_size > len(lru.block):
                lru = ru

        return lru


def detect_repetitive_block_lengths(sequence: str, longest_ref_unit:RepeatUnit|None=None, reverse: bool = False) -> list[RepeatUnit]:
    """Detects the length of repetitive blocks in a string, with an option to search from left to right or reverse.

    In reverse mode, it creates the largest possible blocks of the smallest possible units.
    """
    result: list[RepeatUnit] = []
    seq_len = len(sequence)

    if longest_ref_unit is not None:
        # look for full containment of the longest ref repeat
        # this is so we can detect [2]>[1] (or really any repeat length variation to just 1)
        pattern = f'({re.escape(longest_ref_unit.repeat_unit)})+'
        match = re.search(pattern, sequence)
        
        if match:
            repeat_count = len(sequence) // len(longest_ref_unit.repeat_unit)
            block = repeat_count* longest_ref_unit.repeat_unit
            ru = RepeatUnit(repeat_count, longest_ref_unit.repeat_unit, len(block), block)                        
            result.append(ru)
            shuffleable_bases = sequence[repeat_count*len(ru.repeat_unit):]
            rus = detect_repetitive_block_lengths(shuffleable_bases, reverse = reverse)
            result.extend(rus)
            return result
            

        
    if reverse:
        i = seq_len  # Start from the end of the sequence
        while i > 0:
            matched = False
            # Iterate over block sizes from smallest to largest
            for block_size in range(1, seq_len // 2 + 1):
                if i - block_size < 0:
                    continue  # Not enough characters to form a repeat unit

                # Extract the potential repeat unit ending at position i
                repeat_unit = sequence[i - block_size:i]

                # Calculate the maximum possible number of repeats for this block size
                max_possible_repeats = i // block_size

                # Initialize repeat_count to the maximum possible and decrease
                for repeat_count in range(max_possible_repeats, 1, -1):
                    start_index = i - repeat_count * block_size
                    if start_index < 0:
                        continue  # Not enough characters to form the repetitive block

                    # Extract the substring that could be the repetitive block
                    substr = sequence[start_index:i]

                    # Build the regex pattern for the current repeat unit and count
                    pattern = rf'({re.escape(repeat_unit)})' + r'{' + f'{repeat_count}' + r'}'

                    # Check if the substring matches the pattern
                    if re.fullmatch(pattern, substr):
                        repetitive_block = substr
                        ru = RepeatUnit(
                            block=repetitive_block,
                            block_size=len(repetitive_block),
                            repeat_unit=repeat_unit,
                            repeat_count=repeat_count
                        )
                        result.append(ru)
                        # Move the index `i` backward by the length of the repetitive block
                        i -= len(repetitive_block)
                        matched = True
                        break  # Found the largest block for this unit size

                if matched:
                    break  # Proceed to the next position `i`

            if not matched:
                # No repeat found, remove one character
                ru = RepeatUnit(
                    block=sequence[i - 1],
                    block_size=1,
                    repeat_unit=sequence[i - 1],
                    repeat_count=1
                )
                result.append(ru)
                i -= 1  # Move back by one character if no match is found

    else:
        i = 0  # Start from the beginning of the sequence
        while i < seq_len:
            matched = False
            for block_size in range(1, seq_len // 2 + 1):
                # Build the regex pattern for the current block size
                pattern = rf'(.{{{block_size}}})\1+'
                match = re.match(pattern, sequence[i:])

                if match:
                    repetitive_block = match.group()  # The full repeating pattern
                    repeated_unit = match.group(1)    # The repeating unit
                    repetition_count = len(repetitive_block) // len(repeated_unit)

                    # Add the repetitive block and its details to the result
                    ru = RepeatUnit(
                        block=repetitive_block,
                        block_size=len(repetitive_block),
                        repeat_unit=repeated_unit,
                        repeat_count=repetition_count
                    )
                    result.append(ru)

                    # Move the index `i` forward by the length of the repetitive block
                    i += len(repetitive_block)
                    matched = True
                    break  # Found a match, break the loop for current block size

            if not matched:
                i += 1  # Move forward by one character if no match is found

    return result


def get_repeat_str(
    sequence: str,
    other_seq: str,
    primary_repeat_unit: list[RepeatUnit],
    other_repeat_unit: list[RepeatUnit],
    reverse: bool = False
) -> str:
    if len(primary_repeat_unit) == 0 and len(other_repeat_unit) == 0:
        return None

    if len(primary_repeat_unit) == 0 and len(other_repeat_unit) == 1 and sequence == other_repeat_unit[0].repeat_unit:
        return f"{sequence}[1]"
    elif len(primary_repeat_unit) == 0 and len(other_repeat_unit) == 1 and sequence != other_repeat_unit[0].repeat_unit:
        return None

    if len(primary_repeat_unit) > 0 and len(other_repeat_unit) > 0:
        return_str = assemble_repeat_string(sequence, primary_repeat_unit, reverse=reverse)
        return return_str

    if len(other_repeat_unit) == 0 and len(other_seq) > 0:
        return_str = assemble_repeat_string(sequence, primary_repeat_unit, reverse=reverse)
        if len(return_str) > 0:
            return return_str

    return None

def assemble_repeat_string(sequence: str, repeat_units: list[RepeatUnit], reverse: bool = False) -> str:
    return_str = ""
    primary_repeat_unit = repeat_units.copy()
    seq = sequence
    if reverse:
        while len(seq) > 0:
            found_unit = None
            for ru in primary_repeat_unit:
                if seq.endswith(ru.block):
                    return_str = f"{ru.repeat_unit}[{ru.repeat_count}]" + return_str
                    seq = seq[:-len(ru.block)]
                    found_unit = ru
                    break
            if not found_unit:
                # remove one character if no matching repeat unit is found
                count = 1
                seq_char = seq[0]
                seq = seq[:-1]

                # count consecutive repeating chars  from the end
                while seq and seq[-1] == seq_char:
                    count += 1
                    seq = seq[:-1]

                return_str = f"{seq_char}[{count}]" + return_str

    else: # forward direction
        while len(seq) > 0:
            found_unit = None
            for ru in primary_repeat_unit:
                if seq.startswith(ru.block):
                    return_str += f"{ru.repeat_unit}[{ru.repeat_count}]"
                    seq = seq[len(ru.block):]
                    found_unit = ru
                    break
            if not found_unit:
                # remove one character if no matching repeat unit is found
                count = 1
                seq_char = seq[0]
                seq = seq[1:]

                # count consecutive repeating chars 
                while seq and seq[0] == seq_char:
                    count += 1
                    seq = seq[1:]
                
                return_str += f"{seq_char}[{count}]"

    
    return return_str




"""Shared position handling functions for HGVS."""

import hgvs.location


def get_start_end(
    var, outer_confidence=True
) -> tuple[
    hgvs.location.SimplePosition | hgvs.location.BaseOffsetPosition,
    hgvs.location.SimplePosition | hgvs.location.BaseOffsetPosition,
]:
    """Get start and end positions from a variant or interval.

    This function handles all position types (SimplePosition, BaseOffsetPosition,
    Interval, BaseOffsetInterval) and returns the appropriate start and end positions.
    It can be expected that the returned positions have a base and an uncertain property.

    By default we return the outer confidence positions. However, if that position
    does not have a base, we return the inner confidence positions.

    TODO: add a new optional parameter that allows to define the strictness of the returned positions.
    The current behavior is more alike to an "auto" mode, since we might fall back to the inner confidence positions
    if the outer confidence positions do not have a base. A potential "strict" mode would only return the outer confidence positions,
    and raise an error if the outer confidence positions do not have a base.

    Args:
        var: A variant object with posedit.pos attribute, or an Interval object
        outer_confidence: If True, return the outer confidence positions, otherwise return the inner confidence positions

    Returns:
        tuple: (start_position, end_position) where positions can be SimplePosition or BaseOffsetPosition
    """

    # Handle Interval objects directly
    if isinstance(var, hgvs.location.Interval):
        s = var.start
        e = var.end
        if not isinstance(
            s, (hgvs.location.SimplePosition, hgvs.location.BaseOffsetPosition)
        ):
            if outer_confidence and not s.start.base is not None:
                s = s.start
            else:
                s = s.end
        if not isinstance(
            e, (hgvs.location.SimplePosition, hgvs.location.BaseOffsetPosition)
        ):
            if outer_confidence and not e.end.base is not None:
                e = e.end
            else:
                e = e.start
        return s, e

    # Handle position objects directly
    if isinstance(
        var, (hgvs.location.SimplePosition, hgvs.location.BaseOffsetPosition)
    ):
        return var, var

    # if there is no posedit, return None, all steps below this would fail
    if var.posedit is None:
        return None, None

    # Handle variants with posedit
    pos = var.posedit.pos

    if isinstance(pos, hgvs.location.SimplePosition):
        return pos, pos
    elif isinstance(pos, hgvs.location.BaseOffsetInterval):
        return pos.start, pos.end
    elif isinstance(pos, hgvs.location.Interval):
        s = pos.start
        e = pos.end

        if isinstance(s, hgvs.location.AAPosition) and isinstance(
            e, hgvs.location.AAPosition
        ):
            s = s.base
            e = e.base
            return s, e

        if not isinstance(
            s, (hgvs.location.SimplePosition, hgvs.location.BaseOffsetPosition)
        ):
            orig_s = s

            if outer_confidence and s.start.base is not None:
                s = s.start
            else:
                s = s.end
            if s.is_uncertain:
                s = orig_s.end

        if not isinstance(
            e, (hgvs.location.SimplePosition, hgvs.location.BaseOffsetPosition)
        ):
            orig_e = e
            if outer_confidence and e.end.base is not None:
                e = e.end
            else:
                e = e.start
            if e.is_uncertain:
                e = orig_e.start
        return s, e
    else:  # BaseOffsetPosition
        return pos, pos


def get_start_end_interbase(
    pos: hgvs.location.BaseOffsetPosition | hgvs.location.Interval,
    outer_confidence=True,
) -> tuple[int, int]:
    """Get start and end integer positions from a SequenceVariant.

    This function extracts integer positions from a SequenceVariant, handling uncertain positions
    and intervals. For uncertain positions, it uses the more confident boundary.

    Args:
        var: A SequenceVariant object
        outer_confidence: If True, return the outer confidence positions, otherwise return the inner confidence positions

    Returns:
        tuple: (start_int, end_int) where both are integer positions (0-based)
    """

    # Handle start position
    if pos.start.uncertain:
        # For uncertain start, use the more confident boundary
        if isinstance(pos.start, hgvs.location.Interval):
            if outer_confidence and pos.start.start.base is not None:
                seq_start = pos.start.start.base - 1
            else:
                seq_start = pos.start.end.base - 1
        else:
            seq_start = pos.start.base - 1
    else:
        # For certain start, use the start boundary
        if isinstance(pos.start, hgvs.location.Interval):
            if outer_confidence:
                seq_start = pos.start.start.base - 1
            else:
                seq_start = pos.start.end.base - 1
        else:
            seq_start = pos.start.base - 1

    # Handle end position
    if pos.end.uncertain:
        # For uncertain end, use the more confident boundary
        if isinstance(pos.end, hgvs.location.Interval):
            if outer_confidence and pos.end.end.base is not None:
                seq_end = pos.end.end.base
            else:
                seq_end = pos.end.start.base
        else:
            seq_end = pos.end.base
    else:
        # For certain end, use the end boundary
        if isinstance(pos.end, hgvs.location.Interval):
            if outer_confidence:
                seq_end = pos.end.end.base
            else:
                seq_end = pos.end.start.base
        else:
            seq_end = pos.end.base

    return seq_start, seq_end
